"""
Module :mod:`penta`

This module provides utilities for working with pentadiagonal matrices, namely

- factorisation
- solving
- computation of their log determinant

For accessing the Numba-compatible low-level functions, the submodule :mod:`numba`
is available.

"""

# === Imports ===


